<?xml version="1.0"?>
<!DOCTYPE gui SYSTEM "kpartgui.dtd">
<gui name="kmplayer" version="8">
<MenuBar>
  <Menu name="view"><text>&amp;View</text>
    <Action name="play"/>
    <Action name="pause"/>
    <Action name="stop"/>
    <Action name="edit_playlist_item"/>
  </Menu>
</MenuBar>
<ActionProperties>
  <Action shortcut="F2" name="edit_playlist_item"/>
</ActionProperties>
<ToolBar name="mainToolBar"><text></text>
  <Action name="play"/>
  <Action name="pause"/>
  <Action name="stop"/>
</ToolBar>
</gui>
